// by Alexander Ilyin
// make access to files to be more convenient

#include "file.h"
#include "error.h"
#include <string>
using namespace std;


// *for convenience
LARGE_INTEGER offset_0 = {0};
// *temporary transient macro
#define SA A

// need these 64 bit functions; MSVC 6.0 libraries don't have them
#if _MSC_VER <= 1200
typedef BOOL WINAPI GetFileSizeEx_t (HANDLE hfile, PLARGE_INTEGER file_size);
typedef BOOL WINAPI SetFilePointerEx_t (HANDLE hfile, LARGE_INTEGER distance_to_move, LARGE_INTEGER* new_file_pointer, DWORD move_method);
GetFileSizeEx_t* GetFileSizeEx = 0;
SetFilePointerEx_t* SetFilePointerEx = 0;
#endif
#define CALL_KERNEL 1
// kernel stuff
#if CALL_KERNEL
typedef LONG NTSTATUS;
#define NT_SUCCESS(Status) (((NTSTATUS)(Status)) >= 0)
typedef unsigned long ULONG_PTR;
struct IO_STATUS_BLOCK
{
	union
	{
		NTSTATUS Status;
		PVOID Pointer;
	};
	ULONG_PTR Information;
};
struct FILE_DISPOSITION_INFORMATION
{
	BOOLEAN DeleteFile;
};
enum FILE_INFORMATION_CLASS
{
	FileDispositionInformation = 13
};
typedef NTSTATUS NTAPI NtSetInformationFile_t (HANDLE FileHandle, IO_STATUS_BLOCK* IoStatusBlock, VOID* FileInformation, ULONG Length, FILE_INFORMATION_CLASS FileInformationClass);
NtSetInformationFile_t* NtSetInformationFile = 0;
#endif


static bool InitFunctions ()
{
#if _MSC_VER <= 1200
	HMODULE h_kernel32;

	h_kernel32 = GetModuleHandle (TEXT("KERNEL32"));
	A(h_kernel32);
	*(FARPROC*)&GetFileSizeEx = GetProcAddress (h_kernel32, "GetFileSizeEx");
	if (! GetFileSizeEx)
		return 0;
	*(FARPROC*)&SetFilePointerEx = GetProcAddress (h_kernel32, "SetFilePointerEx");
	if (! SetFilePointerEx)
		return 0;
#endif

#if CALL_KERNEL
	HMODULE h_ntdll;

	h_ntdll = GetModuleHandle (TEXT("NTDLL"));
	A(h_ntdll);
	*(FARPROC*)&NtSetInformationFile = GetProcAddress (h_ntdll, "NtSetInformationFile");
	if (! NtSetInformationFile)
		return 0;
//	NtSetInformationFile = 0;
#endif

	return 1;
}

void File::GetError ()
{
	DWORD D;

	D = GetLastError ();
	switch (D)
	{
	case ERROR_INVALID_HANDLE:
// *this should be some our internal error
		ET;
		break;
	case ERROR_FILE_NOT_FOUND:
		m_error = FILE_ERR__NO_FILE;
		break;
	case ERROR_SHARING_VIOLATION:
	case ERROR_ACCESS_DENIED:
		m_error = FILE_ERR__NO_ACCESS;
		break;
	case ERROR_HANDLE_DISK_FULL:
	case ERROR_DISK_FULL:
		m_error = FILE_ERR__DISK_FULL;
		break;
	}
}

bool File::Open (bool write, bool widechar, const void* filepath)
{
//	BOOL B_r;
	DWORD access;
	DWORD share;
	DWORD disposition;

	m_error = write ? FILE_ERR__UNSPECIFIED_OPEN_FOR_WRITE_ERROR : FILE_ERR__UNSPECIFIED_OPEN_FOR_READ_ERROR;

// not very nice place for that, but it isn't so horrible
//#if _MSC_VER <= 1200
	bool r;
	r = InitFunctions ();
	if (! r)
	{
		return 0;
	}
//#endif

#if 0
// *members are not zeroed by the default constructor
	memset (this, 0, sizeof(*this));
#endif
	A(! m_is_opened);
	A(! m_hfile);
	m_write = write;
	m_pos = 0;

#if 0
// impossible, because it was just zeroed a moment ago ^^
// commander sequence error
	if (m_hfile)
		return 0;
#endif
#if 0
// save the filepath for a possible usage in the future
//	memcpy (m_arc_name, filepath, strlen ((char*)filepath) + 1);
	strncpy (m_arc_name, (char*)filepath, sizeof(m_arc_name) - 1);
	m_arc_name[sizeof(m_arc_name) - 1] = 0;
#endif
//	access = write ? GENERIC_WRITE : GENERIC_READ;
	access = write ? GENERIC_WRITE | DELETE : GENERIC_READ;
// *"add to archive" is never useful; always need to have a new, clean archive
// but maybe we have to overwrite an existing file; so have to share it for read
	share = write ? FILE_SHARE_READ : FILE_SHARE_READ | FILE_SHARE_WRITE;
// don't share the file being read, for write - this will lead to unhandled behaviour
// we don't open it for long period anyway
	share = write ? FILE_SHARE_READ : FILE_SHARE_READ;
//	share = write ? 0 : FILE_SHARE_READ | FILE_SHARE_WRITE;
	disposition = write ? CREATE_ALWAYS : OPEN_EXISTING;
	if (widechar)
// *it also works without prepending "\\?\" at the path, but then the length will be limited to MAX_PATH
	{
		wstring ws;
		ws = L"\\\\?\\";
		ws += (wchar_t*)filepath;
		m_hfile = CreateFileW (ws.c_str(), access, share, 0, disposition, FILE_ATTRIBUTE_NORMAL, 0);
	}
	else
		m_hfile = CreateFileA ((PCSTR)filepath, access, share, 0, disposition, FILE_ATTRIBUTE_NORMAL, 0);
	if (m_hfile == INVALID_HANDLE_VALUE)
//	if (m_hfile == -1)
	{
//// nah, the default error is better
		GetError ();
		return 0;
	}
	m_is_opened = 1;
	m_error = FILE_ERR__NONE;
	return 1;
}

bool File::Close ()
{
	BOOL B_r;

	A(m_is_opened);
	A(m_hfile);
	m_error = FILE_ERR__CLOSE_ERROR;
//	if (! m_hfile)
//		return 0;
	B_r = CloseHandle (m_hfile);
	m_hfile = 0;
	m_is_opened = 0;
	if (! B_r)
		return 0;
	m_error = FILE_ERR__NONE;
	return 1;
}

// delete this opened file
// as it is in 'DeleteFileW()'
bool File::Delete ()
{
//	BOOL B_r;
#if CALL_KERNEL
	NTSTATUS status;
	IO_STATUS_BLOCK io_status_block;
	FILE_DISPOSITION_INFORMATION disposition;
#endif

	A(m_is_opened);
	A(m_hfile);
	A(m_write);

//**
// *I don't think 'FILE_FLAG_DELETE_ON_CLOSE' is necessary
#if CALL_KERNEL
	m_error = FILE_ERR__UNSPECIFIED_DELETE_ERROR;
//	if (! m_hfile)
//		return 0;
// *this function works by name, and this is not what we want
#if 0
	B_r = DeleteFile (m_hfile);
	if (! B_r)
		return 0;
#endif
	disposition.DeleteFile = 1;
	status = NtSetInformationFile (m_hfile, &io_status_block, &disposition, sizeof(disposition), FileDispositionInformation);
	if (NT_SUCCESS(status))
		m_error = FILE_ERR__NONE;
#else
	m_error = 0;
#endif

	return 1;
}

bool File::GetSize ()
{
	BOOL B_r;

//	A(m_is_opened);
	A(m_hfile);
	A(! m_write);
	m_error = FILE_ERR__READ_FAULT;
	SA(sizeof(m_len) == sizeof(LARGE_INTEGER));
	B_r = GetFileSizeEx (m_hfile, (LARGE_INTEGER*)&m_len);
	if (! B_r)
	{
		GetError ();
		return 0;
	}
	m_error = FILE_ERR__NONE;
	return 1;
}

bool File::Timestamp32_To_Filetime64 (u64& filetime, u32 timestamp)
{
//	u64 filetime;
#if 0
// *this algo will work only if count from right after a leap year
	const u32 years_diff = 1970 - 1601;
	const u32 quadruplets = years_diff / 4;
	const u32 years_common_remainder = years_diff % 4;
	const u32 days_per_quadruplet = 365 * 4 + 1;
// leap years are :  1600 and each 4th, except 1700, 1800, 1900
	const u32 days_diff = days_per_quadruplet * quadruplets + 365 * years_common_remainder - 3;
#endif

#if 0
	filetime = (u64)timestamp * TI_SECOND;
//	filetime += (u64)3234576 * 3600 * TI_SECOND;
	filetime += (u64)134774 * 24 * 3600 * TI_SECOND;
//	filetime += (u64)days_diff * 24 * 3600 * TI_SECOND;
#endif
#if 1
	filetime = timestamp + (u64)134774 * 24 * 3600;
	filetime *= TI_SECOND;
#endif
	return 1;
}

bool File::Filetime64_To_Timestamp32 (u32& timestamp, u64 filetime)
{
	u64 val;

	val = filetime / TI_SECOND;
	val = filetime - (u64)134774 * 24 * 3600;
	if (val >= (u64)1 << 32)
	{
		timestamp = 0;
		return 0;
	}
	else
		timestamp = val;
	return 1;
}

bool File::GetTime (FILETIME& ft)
{
	BOOL B_r;

//	A(m_is_opened);
	A(m_hfile);
	A(! m_write);
	m_error = FILE_ERR__READ_FAULT;
	B_r = GetFileTime (m_hfile, 0, 0, &ft);
	if (! B_r)
	{
		GetError ();
		return 0;
	}
	m_error = FILE_ERR__NONE;
	return 1;
}

// takes the number of 100-nanosecond intervals since 1601-01-01 (UTC)
bool File::SetTime (FILETIME& ft)
{
	BOOL B_r;

//	A(m_is_opened);
	A(m_hfile);
	A(m_write);
	m_error = FILE_ERR__WRITE_FAULT;
	B_r = SetFileTime (m_hfile, 0, 0, &ft);
	if (! B_r)
	{
		GetError ();
		return 0;
	}
	m_error = FILE_ERR__NONE;
	return 1;
}

// takes 'time_t', which is the number of seconds since 1970-01-01 00:00 (UTC)
bool File::SetTime (u32 timestamp)
{
	bool r;
	u64 filetime;

	r = Timestamp32_To_Filetime64 (filetime, timestamp);
	if (! r)
		return 0;
	r = SetTime (*(FILETIME*)&filetime);
	if (! r)
		return 0;
	return 1;
}

// *unnecessary
bool File::IsNotEOF (bool& is_not_eof)
{
	BOOL B_r;
	LARGE_INTEGER pos;
	LARGE_INTEGER len;

//	A(m_is_opened);
	A(m_hfile);
// *this operation has no sense for write
	A(! m_write);
	is_not_eof = 0;
	m_error = FILE_ERR__READ_FAULT;
	if (! m_hfile)
		return 0;
// get the current position of file pointer
// *for now could use 'm_pos' instead for that
	B_r = SetFilePointerEx (m_hfile, offset_0, &pos, FILE_CURRENT);
	if (! B_r)
		goto error;
	B_r = GetFileSizeEx (m_hfile, &len);
	if (! B_r)
		goto error;
//	if (pos > len)
	if (pos.QuadPart > len.QuadPart)
		return 0;
	is_not_eof = pos.QuadPart < len.QuadPart;
	m_error = FILE_ERR__NONE;
	return 1;

error:
	GetError ();
	return 0;
}

bool File::SetPointer (bool from_current, u64 pos)
{
	BOOL B_r;
	LARGE_INTEGER pos0;
	LARGE_INTEGER pos1;
	LARGE_INTEGER i;
	DWORD dispos;

	m_error = FILE_ERR__READ_FAULT;

//	A(m_is_opened);
	A(m_hfile);
//// *this operation has no sense for write
//	A(! m_write);
//	if (! m_hfile)
//		return 0;
	dispos = from_current ? FILE_CURRENT : FILE_BEGIN;
	if (from_current)
	{
// get the current position
		B_r = SetFilePointerEx (m_hfile, offset_0, &pos0, FILE_CURRENT);
		if (! B_r)
			return 0;
	}
	i.QuadPart = pos;
	B_r = SetFilePointerEx (m_hfile, i, &pos1, dispos);
	if (! B_r)
		return 0;
	if (from_current)
// relative
	{
		if (pos1.QuadPart != pos0.QuadPart + pos)
			return 0;
		m_pos += pos;
	}
	else
// absolute
	{
		if (pos1.QuadPart != pos)
			return 0;
		m_pos = pos;
	}
	m_error = FILE_ERR__NONE;
	return 1;
}

bool File::Read (void* buf, u32 len)
{
	BOOL B_r;
	DWORD num_read;

//	A(m_is_opened);
	A(m_hfile);
	A(! m_write);
	m_error = FILE_ERR__READ_FAULT;
//	if (! m_hfile)
//		return 0;
	B_r = ReadFile (m_hfile, buf, len, &num_read, 0);
// *is 1, even if the file is empty, or it is EOF ('num_read' then is 0)
	if (! B_r)
		goto error;
	m_pos += num_read;
	if (num_read != len)
	{
//		goto error;
		m_error = FILE_ERR__INCOMPLETE_READ;
		return 0;
	}
	m_error = 0;
	return 1;

error:
	GetError ();
	return 0;
}

bool File::Write (const void* buf, u32 len)
{
	BOOL B_r;
	DWORD num_written;

//	A(m_is_opened);
	A(m_hfile);
	A(m_write);
	m_error = FILE_ERR__WRITE_FAULT;
//	if (! m_hfile)
//		return 0;
	B_r = WriteFile (m_hfile, buf, len, &num_written, 0);
	if (! B_r)
		goto error;
	m_pos += num_written;
	if (num_written != len)
		goto error;
	m_error = 0;
	return 1;

error:
	GetError ();
	return 0;
}

